<?php

namespace Drupal\Tests\aws_cloud\Unit\Plugin;

use Drupal\Tests\UnitTestCase;
use Drupal\Component\Utility\Random;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\aws_cloud\Plugin\AwsCloudServerTemplatePlugin;
use Drupal\cloud_server_template\Entity\CloudServerTemplateInterface;
use Drupal\aws_cloud\Service\AwsEc2ServiceInterface;
use Drupal\aws_cloud\Aws\Ec2\ImageInterface;
use Drupal\aws_cloud\Aws\Ec2\KeyPairInterface;
use Drupal\aws_cloud\Aws\Ec2\SecurityGroupInterface;
use Drupal\aws_cloud\Aws\Ec2\NetworkInterfaceInterface;

/**
 * Tests AWS Cloud Template Plugin.
 *
 * @group AWS Cloud
 */
class AwsCloudServerTemplatePluginTest extends UnitTestCase {

  // Amazon VPC Limits - Security groups per VPC (per region).
  const MAX_SECURITY_GROUPS_COUNT = 500;

  private $plugin;
  private $mockAwsEc2Service;
  private $random;

  protected function setUp() {
    $this->mockAwsEc2Service = $this->getMockBuilder(AwsEc2ServiceInterface::class)
      ->getMock();

    $this->plugin = new AwsCloudServerTemplatePlugin(
      [], '', [],
      $this->mockAwsEc2Service,
      $this->getMockBuilder(Messenger::class)
        ->disableOriginalConstructor()
        ->getMock(),
      $this->getMock(EntityTypeManagerInterface::class)
    );

    $this->random = new Random();
  }

  /**
   * Tests launching a instance.
   */
  public function testLaunch() {
    $random = $this->random;

    // Mock object of Image.
    $mock_image = $this->getMock(ImageInterface::class);
    $image_value_map = [
      ['image_id', (object)['value' => 'ami-' . $random->name(8, TRUE)]],
      ['root_device_type', 'ebs'],
    ];
    $mock_image->expects($this->any())
      ->method('get')
      ->will($this->returnValueMap($image_value_map));

    // Mock object of KeyPair.
    $mock_ssh_key = $this->getMock(KeyPairInterface::class);
    $ssh_key_value_map = [
      ['key_pair_name', (object)['value' => $random->name(8, TRUE)]],
    ];
    $mock_ssh_key->expects($this->any())
      ->method('get')
      ->will($this->returnValueMap($ssh_key_value_map));

    // Mock object of NetworkInterface.
    $mock_network = $this->getMock(NetworkInterfaceInterface::class);
    $network_interface_id = 'eni-' . $random->name(8, TRUE);
    $mock_network->expects($this->any())
      ->method('getNetworkInterfaceId')
      ->will($this->returnValue($network_interface_id));

    $vpc_id = 'vpc-' . $random->name(8, TRUE);
    $field_security_groups_test_cases = $this->createFieldSecurityGroupsTestCases($vpc_id);
    $subnet_id = 'subnet-' . $random->name(8, TRUE);

    // Run test cases.
    foreach ($field_security_groups_test_cases as $field_security_groups_test_case) {
      $mock_template = $this->getMock(CloudServerTemplateInterface::class);
      $template_value_map = [
        ['field_test_only', (object)['value' => '1']],
        ['field_image_id', (object)['entity' => $mock_image]],
        ['field_max_count', (object)['value' => 1]],
        ['field_min_count', (object)['value' => 1]],
        ['field_monitoring', (object)['value' => '1']],
        ['field_instance_type', (object)['value' => $random->name(5, TRUE)]],
        ['field_ssh_key', (object)['entity' => $mock_ssh_key]],
        ['field_kernel_id', (object)['value' => 'aki-' . $random->name(8, TRUE)]],
        ['field_ram', (object)['value' => 'ari-' . $random->name(8, TRUE)]],
        ['field_user_data', (object)['value' => $random->string(32, TRUE)]],
        ['field_availability_zone', (object)['value' => $random->name(7, TRUE)]],
        ['field_security_group', $this->extract_array_item($field_security_groups_test_case, 0)],
        ['field_network', (object)['entity' => $mock_network]],
        ['field_subnet', (object)['value' => $subnet_id]],
        ['field_instance_shutdown_behavior', (object)['value' => '1']],
        ['field_vpc', (object)['value' => $vpc_id]]
      ];
      $mock_template->expects($this->any())
        ->method('get')
        ->will($this->returnValueMap($template_value_map));

      // Assert followings for the first argument of method runInstances,
      // 1. Has key SecurityGroups,
      // 2. Doesn't have key SecurityGroup,
      // 3. Contains the array of group1 and group2,
      // 4. Has key SubnetId,
      // 5. Contains subnet_id.
      $this->mockAwsEc2Service
        ->expects($this->any())
        ->method('runInstances')
        ->with($this->logicalAnd(
          $this->arrayHasKey('SecurityGroupIds'),
          $this->logicalNot($this->arrayHasKey('SecurityGroups')),
          $this->contains($this->extract_array_item($field_security_groups_test_case, 1)),
          $this->arrayHasKey('SubnetId'),
          $this->contains($subnet_id)
        ))
        ->will($this->returnValue(TRUE));

      $return = $this->plugin->launch(
        $mock_template,
        $this->getMock(FormStateInterface::class)
      );
      $this->assertNotNull($return);
    }

  }

  private function extract_array_item($array, $index) {
    $extracted = [];
    foreach ($array as $item) {
      $extracted[] = $item[$index];
    }

    return $extracted;
  }

  private function createFieldSecurityGroupsTestCases($vpc_id) {
    $random = $this->random;

    $mock_security_group_and_names = [];
    for ($i = 0; $i < self::MAX_SECURITY_GROUPS_COUNT; $i++) {
      // Mock object of SecurityGroup.
      $mock_security_group = $this->getMock(SecurityGroupInterface::class);
      $group_id = 'sg-' . $random->name(8, TRUE);
      $mock_security_group->expects($this->any())
        ->method('getGroupId')
        ->will($this->returnValue($group_id));

      $mock_security_group->expects($this->any())
        ->method('getVpcId')
        ->will($this->returnValue($vpc_id));

      $mock_security_group_and_ids[] = [
        (object)['entity' => $mock_security_group],
        $group_id
      ];
    }

    $mock_security_group_and_ids_exclude_first_last = array_slice(
      $mock_security_group_and_ids,
      1,
      count($mock_security_group_and_ids) - 2
    );

    $field_security_groups_test_cases = [];

    // Test cases.
    //
    // Test when Security Groups has the only one item, the first item only.
    // Security Group A (* SELECTED).
    // ...
    // Security Group M.
    // ...
    // Security Group N.
    $field_security_groups_test_cases[] = [reset($mock_security_group_and_ids)];

    // Test when Security Groups has more than two items.
    if (self::MAX_SECURITY_GROUPS_COUNT > 1) {

      // The last item only.
      // Security Group A.
      // ...
      // Security Group M.
      // ...
      // Security Group N (* SELECTED).
      $field_security_groups_test_cases[] = [end($mock_security_group_and_ids)];

      // The first item and the last one only.
      // Security Group A (* SELECTED).
      // ...
      // Security Group M.
      // ...
      // Security Group N (* SELECTED).
      $field_security_groups_test_cases[] = [
        reset($mock_security_group_and_ids),
        end($mock_security_group_and_ids)
      ];

      if (self::MAX_SECURITY_GROUPS_COUNT > 2) {

        // Randomly picking up the only one item.

        // One random item.
        // Security Group A.
        // ...
        // Security Group M (* SELECTED).
        // ...
        // Security Group N.
        $field_security_groups_test_cases[] = [
          // Pickup one index randomly (array index: 1 to the (last index -2))
          $mock_security_group_and_ids_exclude_first_last[
            array_rand($mock_security_group_and_ids_exclude_first_last)
          ]
        ];

        // The first item and one random one.
        // Security Group A (* SELECTED).
        // ...
        // Security Group M (* SELECTED).
        // ...
        // Security Group N.
        $field_security_groups_test_cases[] = [
          reset($mock_security_group_and_ids),
          // Pickup one index randomly (array index: 1 to the (last index -2))
          $mock_security_group_and_ids_exclude_first_last[
            array_rand($mock_security_group_and_ids_exclude_first_last)
          ],
        ];

        // One random item and the last one.
        // Security Group A.
        // ...
        // Security Group M (* SELECTED).
        // ...
        // Security Group N (* SELECTED).
        $field_security_groups_test_cases[] = [
          // Pickup one index randomly (array index: 1 to the (last index -2))
          $mock_security_group_and_ids_exclude_first_last[
            array_rand($mock_security_group_and_ids_exclude_first_last)
          ],
          end($mock_security_group_and_ids),
        ];

        if (self::MAX_SECURITY_GROUPS_COUNT > 3) {

          // Two random items.
          // Security Group A.
          // ...
          // Security Group M1 (* SELECTED).
          // ...
          // Security Group M2 (* SELECTED).
          // ...
          // Security Group N.
          $array_indice = array_rand($mock_security_group_and_ids_exclude_first_last, 2);
          $field_security_groups_test_cases[] = [
            $mock_security_group_and_ids_exclude_first_last[
              $array_indice[0]
            ],
            $mock_security_group_and_ids_exclude_first_last[
              $array_indice[1]
            ],
          ];
        }
      }
    }

    return $field_security_groups_test_cases;
  }

}
